@@ -36,7 +36,6 @@ from shapely.geometry import Polygon
 from shapely.wkt import loads as shply_loads
 from six import PY3
 
-# mpl_toolkits.basemap -> imported when GeoArray.show_map() is used
 # dill -> imported when dumping GeoArray
 
 try:
@@ -664,17 +663,21 @@ class GeoArray(object):
 
         self.__dict__ = state
 
-    def calc_mask_nodata(self, fromBand=None, overwrite=False):
-        """Calculates a no data mask with (values: 0=nodata; 1=data)
+    def calc_mask_nodata(self, fromBand=None, overwrite=False, flag='all'):
+        # type: (int, bool, str) -> np.ndarray
+        """Calculates a no data mask with values False (=nodata) and True (=data).
 
-        NOTE:   Only pixel containing the nodata values in ALL bands are recognized as nodata pixel. If they contain a
-                pixel value different from the nodata value in any band they are good data pixels.
-
-        :param fromBand:  <int> index of the band to be used (if None, all bands are used)
-        :param overwrite: <bool> whether to overwrite existing nodata mask that has already been calculated
+        :param fromBand:   index of the band to be used (if None, all bands are used)
+        :param overwrite:  whether to overwrite existing nodata mask that has already been calculated
+        :param flag:       algorithm how to flag pixels (default: 'all')
+                           'all': flag those pixels as nodata that contain the nodata value in ALL bands
+                           'any': flag those pixels as nodata that contain the nodata value in ANY band
         :return:
         """
         if self._mask_nodata is None or overwrite:
+            if flag not in ['all', 'any']:
+                raise ValueError(flag)
+
             assert self.ndim in [2, 3], "Only 2D or 3D arrays are supported. Got a %sD array." % self.ndim
             arr = self[:, :, fromBand] if self.ndim == 3 and fromBand is not None else self[:]
 
@@ -690,21 +693,42 @@ class GeoArray(object):
                 elif arr.ndim == 2:
                     mask = ~np.isnan(arr)
                 else:
-                    idx_1st_databand = np.argwhere(~nanbands)[0][0]
-                    mask = ~np.isnan(arr[:, :, idx_1st_databand])
-                    mask[~mask] = np.any(~np.isnan(arr[~mask]), axis=1)
+                    arr_1st_databand = arr[:, :, np.argwhere(~nanbands)[0][0]]
+                    arr_remain = arr[:, :, ~nanbands][:, :, 1:]
+
+                    mask = ~np.isnan(arr_1st_databand)  # True where 1st data band has data
+
+                    if flag == 'all':
+                        # ALL bands need to contain np.nan to flag the mask as nodata
+                        # overwrite the mask at nodata positions (False) with True in case there is data in ANY band
+                        mask[~mask] = np.any(~np.isnan(arr_remain[~mask]), axis=1)
+                    else:
+                        # ANY band needs to contain np.nan to flag the mask as nodata
+                        # overwrite the mask at data positions (True) with False in case there is np.nan in ANY band
+                        mask[mask] = ~np.any(np.isnan(arr_remain[mask]), axis=1)
 
             else:
                 bandmeans = np.mean(np.mean(arr, axis=0), axis=0)
+                nodatabands = bandmeans == self.nodata
 
                 if np.nanmean(bandmeans) == self.nodata:
                     mask = np.full(arr.shape[:2], False)
                 elif arr.ndim == 2:
                     mask = arr != self.nodata
                 else:
-                    idx_1st_databand = np.argwhere(bandmeans != self.nodata)[0][0]
-                    mask = np.array(arr[:, :, idx_1st_databand] != self.nodata)
-                    mask[~mask] = np.any(arr[~mask] != self.nodata, axis=1)
+                    arr_1st_databand = arr[:, :, np.argwhere(~nodatabands)[0][0]]
+                    arr_remain = arr[:, :, ~nodatabands][:, :, 1:]
+
+                    mask = np.array(arr_1st_databand != self.nodata)  # True where 1st data band has data
+
+                    if flag == 'all':
+                        # ALL bands need to contain nodata to flag the mask as such
+                        # overwrite the mask at nodata positions (False) with True in case there is data in ANY band
+                        mask[~mask] = np.any(arr_remain[~mask] != self.nodata, axis=1)
+                    else:
+                        # ANY band needs to contain nodata to flag the mask as such
+                        # overwrite the mask at data positions (True) with False in case there is nodata in ANY band
+                        mask[mask] = ~np.any(arr_remain[mask] == self.nodata, axis=1)
 
             self.mask_nodata = mask
 
@@ -1106,7 +1130,7 @@ class GeoArray(object):
             # sample image down / normalize
             xdim, ydim = \
                 (self.columns * res_factor, self.rows * res_factor) if res_factor else \
-                tuple((np.array([self.columns, self.rows]) / (np.array([self.columns, self.rows]).max() / 1000)))
+                tuple(np.array([self.columns, self.rows]) / (max([self.columns, self.rows]) / 1000))
             xdim, ydim = int(xdim), int(ydim)
 
         if xdim or ydim or out_prj:
@@ -1128,6 +1152,28 @@ class GeoArray(object):
 
         return image2plot, gt, prj
 
+    @staticmethod
+    def _get_cmap_vmin_vmax(cmap, vmin, vmax, pmin, pmax, image2plot, nodataVal):
+        from matplotlib import pyplot as plt
+        # set color palette
+        palette = plt.get_cmap(cmap) if cmap and isinstance(cmap, str) else cmap if cmap else plt.get_cmap('gray')
+        palette = copy(palette)  # allows to modify the colormap as in-place modifications are not allowed anymore
+        if nodataVal is not None and np.std(image2plot) != 0:  # do not show nodata
+            image2plot = np.ma.masked_equal(image2plot, nodataVal)
+            vmin_auto, vmax_auto = \
+                np.nanpercentile(image2plot.compressed(), pmin), np.nanpercentile(image2plot.compressed(), pmax)
+            palette.set_bad('aqua', 0)
+        else:
+            vmin_auto, vmax_auto = np.nanpercentile(image2plot, pmin), np.nanpercentile(image2plot, pmax)
+
+        vmin = vmin if vmin is not None else vmin_auto
+        vmax = vmax if vmax is not None else vmax_auto
+
+        palette.set_over('1')
+        palette.set_under('0')
+
+        return palette, vmin, vmax
+
     def show(self, xlim=None, ylim=None, band=None, boundsMap=None, boundsMapPrj=None, figsize=None,
              interpolation='none', vmin=None, vmax=None, pmin=2, pmax=98, cmap=None, nodataVal=None,
              res_factor=None, interactive=False, ax=None, ignore_rotation=False):
@@ -1166,22 +1212,9 @@ class GeoArray(object):
             self._get_plottable_image(xlim, ylim, band, boundsMap=boundsMap, boundsMapPrj=boundsMapPrj,
                                       res_factor=res_factor, nodataVal=nodataVal, ignore_rotation=ignore_rotation)
 
-        # set color palette
-        palette = plt.get_cmap(cmap) if cmap and isinstance(cmap, str) else cmap if cmap else plt.get_cmap('gray')
-        palette = copy(palette)  # allows to modify the colormap as in-place modifications are not allowed anymore
-        if nodataVal is not None and np.std(image2plot) != 0:  # do not show nodata
+        palette, vmin, vmax = self._get_cmap_vmin_vmax(cmap, vmin, vmax, pmin, pmax, image2plot, nodataVal)
+        if nodataVal is not None and np.std(image2plot) != 0:
             image2plot = np.ma.masked_equal(image2plot, nodataVal)
-            vmin_auto, vmax_auto = \
-                np.nanpercentile(image2plot.compressed(), pmin), np.nanpercentile(image2plot.compressed(), pmax)
-            palette.set_bad('aqua', 0)
-        else:
-            vmin_auto, vmax_auto = np.nanpercentile(image2plot, pmin), np.nanpercentile(image2plot, pmax)
-
-        vmin = vmin if vmin is not None else vmin_auto
-        vmax = vmax if vmax is not None else vmax_auto
-
-        palette.set_over('1')
-        palette.set_under('0')
 
         # check availability of holoviews
         if interactive and not find_loader('holoviews'):
@@ -1197,7 +1230,8 @@ class GeoArray(object):
             cS, cE = xlim if isinstance(xlim, (tuple, list)) else (0, self.columns - 1)
             rS, rE = ylim if isinstance(ylim, (tuple, list)) else (0, self.rows - 1)
 
-            image2plot = np.array(rescale_intensity(image2plot, in_range=(vmin, vmax)))
+            # noinspection PyTypeChecker
+            image2plot = rescale_intensity(image2plot, in_range=(vmin, vmax))  # type: np.ndarray
 
             def get_hv_image(b):
                 # FIXME ylabels have the wrong order
@@ -1227,9 +1261,9 @@ class GeoArray(object):
 
             return im
 
-    def show_map(self, xlim=None, ylim=None, band=0, boundsMap=None, boundsMapPrj=None, ax=None, figsize=None,
-                 interpolation='none', vmin=None, vmax=None, cmap=None, nodataVal=None, res_factor=None,
-                 return_map=False, zoomable=False):
+    def show_map(self, xlim=None, ylim=None, band=0, boundsMap=None, boundsMapPrj=None, out_epsg=None, figsize=None,
+                 interpolation='none', vmin=None, vmax=None, pmin=2, pmax=98, cmap=None, nodataVal=None,
+                 res_factor=None, return_map=False):
         """
 
         :param xlim:
@@ -1237,42 +1271,22 @@ class GeoArray(object):
         :param band:            band index (starting with 0)
         :param boundsMap:       xmin, ymin, xmax, ymax
         :param boundsMapPrj:
-        :param ax:              allows to pass a matplotlib axis object where figure is plotted into
+        :param out_epsg:        EPSG code of the output projection
         :param figsize:
         :param interpolation:
-        :param vmin:
-        :param vmax:
+        :param vmin:            darkest pixel value to be included in stretching
+        :param vmax:            brightest pixel value to be included in stretching
+        :param pmin:            percentage to be used for excluding the darkest pixels from stretching (default: 2)
+        :param pmax:            percentage to be used for excluding the brightest pixels from stretching (default: 98)
         :param cmap:
         :param nodataVal:
         :param res_factor:      <float> resolution factor for downsampling of the image to be plotted in order to save
                                 plotting time and memory (default=None -> downsampling is performed to 1000x1000)
         :param return_map:
-        :param zoomable:        <bool> enable or disable zooming via mpld3
         :return:
         """
-
-        if not find_loader('mpl_toolkits.basemap'):
-            raise ImportError('This function requires Basemap. You need to install basemap manually (see www./'
-                              'matplotlib.org/basemap) if you want to plot maps. It is not automatically installed.')
-
-        from mpl_toolkits.basemap import Basemap
         from matplotlib import pyplot as plt
-
-        mpld3_avl = find_loader('mpld3')
-        if not mpld3_avl:
-            warnings.warn('mpld3 is not available. Zooming disabled.')
-            zoomable = False
-
-        if zoomable:
-            import mpld3
-            mpld3.enable_notebook()
-        elif mpld3_avl:
-            import mpld3
-            # noinspection PyBroadException
-            try:
-                mpld3.disable_notebook()
-            except Exception:
-                pass
+        from cartopy.crs import epsg as ccrs_from_epsg, PlateCarree
 
         assert self.geotransform and tuple(self.geotransform) != (0, 1, 0, 0, 0, -1), \
             'A valid geotransform is needed for a map visualization. Got %s.' % list(self.geotransform)
@@ -1282,126 +1296,38 @@ class GeoArray(object):
         nodataVal = nodataVal if nodataVal is not None else self.nodata
         image2plot, gt, prj = self._get_plottable_image(xlim, ylim, band, boundsMap=boundsMap,
                                                         boundsMapPrj=boundsMapPrj, res_factor=res_factor,
-                                                        nodataVal=nodataVal, out_prj='epsg:4326')
-
-        # calculate corner coordinates of plot
-        # if boundsMap:
-        #    boundsMapPrj = boundsMapPrj if boundsMapPrj else self.prj
-        #    if not prj_equal(boundsMapPrj, 4326):
-        #        boundsMap = reproject_shapelyGeometry(box(*boundsMap), boundsMapPrj, 4626).bounds
-        #    xmin, ymin, xmax, ymax = boundsMap
-        #    UL_XY, UR_XY, LR_XY, LL_XY = (xmin,ymax), (xmax, ymax), (xmax,ymin), (xmin, ymin)
-        # else:
-        UL_XY, UR_XY, LR_XY, LL_XY = [(YX[1], YX[0]) for YX in GeoArray(image2plot, gt, prj).box.boxMapYX]
-        center_lon, center_lat = (UL_XY[0] + UR_XY[0]) / 2., (UL_XY[1] + LL_XY[1]) / 2.
+                                                        nodataVal=nodataVal)
 
         # create map
+        def get_cartopy_crs_from_epsg(epsg_code):
+            try:
+                return ccrs_from_epsg(epsg_code)
+            except ValueError:
+                if epsg_code == 4326:
+                    return PlateCarree()
+                else:
+                    raise NotImplementedError('The show_map() method currently does not support the given projection.')
+
+        crs_in = get_cartopy_crs_from_epsg(self.epsg)
+        crs_out = get_cartopy_crs_from_epsg(out_epsg if out_epsg is not None else self.epsg)
+
         fig = plt.figure(figsize=figsize)
-        plt.subplots_adjust(left=0.05, right=0.95, top=0.90, bottom=0.05, wspace=0.15, hspace=0.05)
-        ax = ax if ax is not None else plt.subplot(111)
+        ax = plt.axes(projection=crs_out)
 
-        m = Basemap(projection='tmerc', ellps='WGS84', resolution=None, lon_0=center_lon, lat_0=center_lat,
-                    urcrnrlon=UR_XY[0], urcrnrlat=UR_XY[1], llcrnrlon=LL_XY[0], llcrnrlat=LL_XY[1])
+        ax.set_extent(self.box.boundsMap, crs=crs_in)
 
-        # set color palette
-        palette = cmap if cmap else plt.get_cmap('gray')
+        palette, vmin, vmax = self._get_cmap_vmin_vmax(cmap, vmin, vmax, pmin, pmax, image2plot, nodataVal)
         if nodataVal is not None and np.std(image2plot) != 0:  # do not show nodata
             image2plot = np.ma.masked_equal(image2plot, nodataVal)
-            vmin_auto, vmax_auto = \
-                np.nanpercentile(image2plot.compressed(), 2), np.nanpercentile(image2plot.compressed(), 98)
-            palette.set_bad('aqua', 0)
-        else:
-            vmin_auto, vmax_auto = np.nanpercentile(image2plot, 2), np.nanpercentile(image2plot, 98)
-        vmin = vmin if vmin is not None else vmin_auto
-        vmax = vmax if vmax is not None else vmax_auto
-        palette.set_over('1')
-        palette.set_under('0')
+        ax.imshow(image2plot, cmap=palette, interpolation=interpolation, vmin=vmin, vmax=vmax,
+                  origin='upper', transform=crs_in, extent=list(self.box.boundsMap))
 
-        # add image to map (y-axis must be inverted for basemap)
-        if zoomable:
-            m.imshow(image2plot, palette, interpolation=interpolation, vmin=vmin, vmax=vmax)
-        else:
-            m.imshow(np.flipud(image2plot), palette, interpolation=interpolation, vmin=vmin, vmax=vmax)
-
-        # add coordinate grid lines
-        parallels = np.arange(-90, 90., 0.25)  # TODO make this adjustable
-        # parallels = np.arange(-90, 90., 0.1)
-        m.drawparallels(parallels, labels=[1, 0, 0, 0], fontsize=12, linewidth=0.4)
-
-        meridians = np.arange(-180., 180., 0.25)
-        # meridians = np.arange(-180., 180., 0.1)
-        m.drawmeridians(meridians, labels=[0, 0, 0, 1], fontsize=12, linewidth=0.4)
+        # show grid lines
+        # NOTE: the current version of cartopy can only label PlateCarree grid lines - (crs=crs) is not yet supported
+        ax.gridlines(draw_labels=True, linewidth=2, color='gray', alpha=0.5, linestyle='--')
 
         if return_map:
-            return fig, ax, m
-        else:
-            plt.show()
-
-    def show_map_utm(self, xlim=None, ylim=None, band=0, figsize=None, interpolation='none', cmap=None,
-                     nodataVal=None, vmin=None, vmax=None, res_factor=None, return_map=False):
-
-        if not find_loader('mpl_toolkits.basemap'):
-            raise ImportError('This function requires Basemap. You need to install basemap manually (see www./'
-                              'matplotlib.org/basemap) if you want to plot maps. It is not automatically installed.')
-
-        from mpl_toolkits.basemap import Basemap
-        from matplotlib import pyplot as plt
-
-        warnings.warn(UserWarning('This function is still under construction and may not work as expected!'))
-        # TODO debug this function
-
-        # get image to plot
-        nodataVal = nodataVal if nodataVal is not None else self.nodata
-        image2plot, gt, prj = self._get_plottable_image(xlim, ylim, band, res_factor, nodataVal)
-
-        # calculate corner coordinates of plot
-        box2plot = GeoArray(image2plot, gt, prj).box
-        # UL_XY, UR_XY, LR_XY, LL_XY = [(YX[1], YX[0]) for YX in GeoArray(image2plot, gt, prj).box.boxMapYX]
-        # Xarr, Yarr = self.box.get_coordArray_MapXY(prj=EPSG2WKT(4326))
-        UL_XY, UR_XY, LR_XY, LL_XY = [transform_any_prj(self.projection, 'epsg:4326', x, y) for y, x in
-                                      box2plot.boxMapYX]
-        center_X, center_Y = (UL_XY[0] + UR_XY[0]) / 2., (UL_XY[1] + LL_XY[1]) / 2.
-        center_lon, center_lat = transform_any_prj(prj, 'epsg:4326', center_X, center_Y)
-        print(center_lon, center_lat)
-
-        # create map
-        fig = plt.figure(figsize=figsize)
-        plt.subplots_adjust(left=0.05, right=0.95, top=0.90, bottom=0.05, wspace=0.15, hspace=0.05)
-        ax = plt.subplot(111)
-        print(UL_XY, UR_XY, LR_XY, LL_XY)
-        #        m = Basemap(projection='tmerc', resolution=None, lon_0=center_lon, lat_0=center_lat,
-        #                    urcrnrx=UR_XY[0], urcrnry=UR_XY[1], llcrnrx=LL_XY[0], llcrnry=LL_XY[1])
-        m = Basemap(projection='tmerc', ellps='WGS84', resolution=None, lon_0=center_lon, lat_0=center_lat,
-                    urcrnrlon=UR_XY[0], urcrnrlat=UR_XY[1], llcrnrlon=LL_XY[0], llcrnrlat=LL_XY[1],
-                    k_0=0.9996, rsphere=(6378137.00, 6356752.314245179), suppress_ticks=False)
-        # m.pcolormesh(Xarr, Yarr, self[:], cmap=plt.cm.jet)
-
-        # set color palette
-        palette = cmap if cmap else plt.get_cmap('gray')
-        if nodataVal is not None:  # do not show nodata
-            image2plot = np.ma.masked_equal(image2plot, nodataVal)
-            vmin_auto, vmax_auto = \
-                np.nanpercentile(image2plot.compressed(), 2), np.nanpercentile(image2plot.compressed(), 98)
-            palette.set_bad('aqua', 0)
-        else:
-            vmin_auto, vmax_auto = np.nanpercentile(image2plot, 2), np.nanpercentile(image2plot, 98)
-        vmin = vmin if vmin is not None else vmin_auto
-        vmax = vmax if vmax is not None else vmax_auto
-        palette.set_over('1')
-        palette.set_under('0')
-
-        # add image to map (y-axis must be inverted for basemap)
-        m.imshow(np.flipud(image2plot), palette, interpolation=interpolation, vmin=vmin, vmax=vmax)
-
-        # add coordinate grid lines
-        parallels = np.arange(-90, 90., 0.25)
-        m.drawparallels(parallels, labels=[1, 0, 0, 0], fontsize=12, linewidth=0.4)
-
-        meridians = np.arange(-180., 180., 0.25)
-        m.drawmeridians(meridians, labels=[0, 0, 0, 1], fontsize=12, linewidth=0.4)
-
-        if return_map:
-            return fig, ax, m
+            return fig, ax
         else:
             plt.show()
 
